use crate::background::Background;
use crate::maps::Maps;
use crate::rgb::Rgb;
use crate::sprite::Sprite;
use crate::tilemap::Tilemap;
use crate::{
    Buttons, BACKGROUND_MAX_SIZE, DPAD_DOWN, DPAD_LEFT, DPAD_RIGHT, DPAD_UP, EMPTY_TILE,
    MAX_BACKGROUNDS, MAX_SPRITES,
};
use ascii::{AsciiChar, AsciiStr};
use glam::IVec2;

/// Complete state of GSA
pub struct Gsa {
    /// Sprites available
    pub sprite: [Sprite; MAX_SPRITES],

    /// Palette used to draw graphics, initially loaded from gfx.gif
    pub palette: [Rgb; 256],

    /// Tilemap layers available, layer 0 in the back, layer 3 in front; layer 3 defaults to half-tile mode
    pub bg: [Background; MAX_BACKGROUNDS],

    /// Currently selected font
    ///
    /// Chosen as half-size tile index, extends 16x16 half tiles in x and y
    pub font: u16,

    /// Target BG for string functions, defaults to layer 3
    pub str_bg: usize,

    pub(crate) pressed: Buttons,
    pub(crate) released: Buttons,
    pub(crate) down: Buttons,
    pub(crate) maps: Maps,
}

impl Gsa {
    /// Resets all bg
    pub fn reset_bgs(&mut self) {
        for bg in 0..MAX_BACKGROUNDS {
            self.bg[bg].clear();
            self.bg[bg].scroll = IVec2::ZERO;
            self.bg[bg].half_tile = false;
        }
        //todo: document ui layer defaulting to half tile
        self.bg[3].half_tile = true;
    }

    /// Clears all sprites
    pub fn reset_sprites(&mut self) {
        for sprite in 0..MAX_SPRITES {
            self.sprite[sprite].pos = IVec2::ZERO;
            self.sprite[sprite].tile = EMPTY_TILE;
            self.sprite[sprite].priority = 0;
        }
    }

    /// Loads tilemap into backgrounds from id
    pub fn load_map(&mut self, map: u16) {
        for (i, map) in self.maps.maps[&map].iter().enumerate() {
            self.bg[i].size = map.size;
            for y in 0..map.size.y as usize {
                for x in 0..map.size.x as usize {
                    self.bg[i].tiles[x][y] = map.data[x + y * map.size.x as usize]
                }
            }
        }
    }

    /// Stores tilemap from backgrounds into id
    pub fn store_map(&mut self, map: u16) {
        self.maps.maps.insert(
            map,
            [
                Tilemap::from_bg(&self.bg[0]),
                Tilemap::from_bg(&self.bg[1]),
                Tilemap::from_bg(&self.bg[2]),
            ],
        );
    }

    /// Does a tilemap with this id exist?
    pub fn map_exists(&mut self, map: u16) -> bool {
        self.maps.maps.contains_key(&map)
    }

    /// Checks if any of given buttons currently held down
    pub fn button_down(&self, button: Buttons) -> bool {
        self.down & button != 0
    }

    /// Checks if any of given buttons been pressed in the current frame
    pub fn button_pressed(&self, button: Buttons) -> bool {
        self.pressed & button != 0
    }

    /// Checks if any of given buttons been released in the current frame
    pub fn button_released(&self, button: Buttons) -> bool {
        self.released & button != 0
    }

    /// Directional vector (-1 to 1) from current dpad state
    pub fn input_dir(&self) -> IVec2 {
        IVec2 {
            x: if self.button_down(DPAD_LEFT) { -1 } else { 0 }
                + if self.button_down(DPAD_RIGHT) { 1 } else { 0 },
            y: if self.button_down(DPAD_UP) { -1 } else { 0 }
                + if self.button_down(DPAD_DOWN) { 1 } else { 0 },
        }
    }

    /// Write given string on given map, at given position, with font [Gsa::font]
    pub fn write_string(&mut self, pos: IVec2, str: &str) {
        let str = AsciiStr::from_ascii(str).unwrap();
        for (i, ch) in str.into_iter().enumerate() {
            self.bg[self.str_bg].tiles[pos.x as usize + i][pos.y as usize] =
                self.get_char_tile(*ch);
        }
    }

    /// Write given string vertically on given map, at given position, with font [Gsa::font]
    pub fn write_string_vertical(&mut self, pos: IVec2, str: &str) {
        let str = AsciiStr::from_ascii(str).unwrap();
        for (i, ch) in str.into_iter().enumerate() {
            self.bg[self.str_bg].tiles[pos.x as usize][pos.y as usize + i] =
                self.get_char_tile(*ch);
        }
    }

    fn get_char_tile(&self, ch: AsciiChar) -> u16 {
        let ch = ch as u16;
        self.font + (ch % 0x10) + (ch / 0x10) * 0x100
    }
}
